"""Module for investigation of thread activity.
Runs with some examples.

"""
import os
import threading
from time import sleep


class Color:
    PURPLE = '\033[1;35;48m'
    CYAN = '\033[1;36;48m'
    BOLD = '\033[1;37;48m'
    BLUE = '\033[1;34;48m'
    GREEN = '\033[1;32;48m'
    YELLOW = '\033[1;33;48m'
    RED = '\033[1;31;48m'
    BLACK = '\033[1;30;48m'
    UNDERLINE = '\033[4;37;48m'
    END = '\033[1;37;0m'


color = Color()


def mp_start_show_threads(toolbox):
    """"""
    pid = os.getpid()
    threading.Thread(name='Watchdog_' + str(pid),
                     target=mp_show_threads,
                     args=(toolbox,),
                     daemon=True).start()


def mp_show_threads(toolbox):
    """Display running thread names.

    :params: toolbox: queues and vars
    """
    toolbox.mp_print_q.put(f' .. thread display interval {toolbox.sleep_time}s')
    while 1:
        names = [thread.name for thread in threading.enumerate()]
        sorted_ = sorted(names)
        msg = color.CYAN + f'process name: {toolbox.WORKER_NAME} pid {toolbox.WORKER_PID} {sorted_}' + color.END
        toolbox.mp_print_q.put(msg)
        sleep(toolbox.sleep_time)


def show_threads(sleep_time):
    """Display running thread names.

    :params: sleep_time: display interval
    """
    print(f' .. thread display interval {sleep_time}s')
    while 1:
        names = [thread.name for thread in threading.enumerate()]
        sorted_ = sorted(names)
        print(sorted_)
        sleep(sleep_time)



"""Green CSV generator / Worker in folder /worker

We download a NZ government report and calculate the average of split lists.
See URL. Try your own URL. They have a lot.
"""

import os
import csv
import time
import platform
from io import TextIOWrapper
from zipfile import ZipFile

import eisenmp
import eisenmp.utils.eisenmp_utils as e_utils
try:
    from eisenmp.utils_exa.eisenmp_download import DownLoad
except ImportError:
    from eisenmp_examples.utils_exa.eisenmp_download import DownLoad

dir_name = os.path.dirname(__file__)


class ModuleConfiguration:
    """Fill the existing attributes.
    You can use the class to have your variables available in the module.
    You have full access to all queues and methods. mp = eisenmp.Mp()

    """
    # path to worker module and entry function reference, worker module import in
    # [isolated, dead references to data structures, [list, dict, instance], offset addr. unreachable] process env
    # -------------------- MANDATORY WORKER STRINGS (large worker live in /worker) --------------------
    first_module = {
        'worker_path': os.path.join(dir_name, 'worker', 'eisenmp_exa_wrk_csv.py'),
        'worker_ref': 'worker_entrance',
    }
    watchdog_module = {
        'WORKER_PATH': os.path.join(dir_name, 'worker', 'eisenmp_exa_wrk_watchdog.py'),
        'WORKER_REF': 'mp_start_show_threads',
    }
    dl_url = 'https://www.stats.govt.nz/assets/Uploads/International-trade/' \
             'International-trade-December-2022-quarter/Download-data/international-trade-december-2022-quarter.zip'

    def __init__(self):
        # load order list, first module is called in an endless loop, you can append your own loop inside the worker
        self.worker_modules = [
            self.first_module,  # second module must be started by a thread, else we hang
            self.watchdog_module,
        ]

        # Multiprocess vars - override default
        self.NUM_PROCS = 5  # your process count, default is None: one proc/CPU core
        self.NUM_ROWS = 50_000  # workload spread, list (generator items) to calc in one loop, default is None: 1_000
        self.RESULTS_STORE = True  # keep in dictionary, will crash the system if store GB network chunks in mem
        self.RESULTS_PRINT = True  # result rows of output are collected in a list, display if processes are stopped
        self.RESULTS_DICT_PRINT = True  # shows content of results dict with ticket numbers, check tickets
        # max generator / NUM_ROWS = number of tickets, 10_000 / 42 = 238.095 -> 238 lists with ticket numbers
        self.RESULT_LABEL = 'revised.csv, Average calculation'  # pretty print as result header for RESULTS_PRINT
        if platform.system() == 'Linux':
            self.START_METHOD = 'fork'  # 'spawn' is default if unused; also use 'forkserver' or 'fork' on Unix only

        # CSV part
        self.use_file_system = False  # False: download and unzip in mem; True must exist on fs ------------- SWITCH ---
        self.url = self.dl_url  # False 'use_file_system', URL of csv file
        self.zipped_filename = 'revised.csv'  # name of the uncompressed file in zip archive
        self.csv_col_name = 'value'  # CSV table column header
        self.sleep_time = 45  # watchdog module in 'worker_modules' list


modConf = ModuleConfiguration()  # accessible in module


def generator_calc_csv():
    """
    - Manager -

    Generator part of calc CSV. Worker module in /worker folder.

    :params: result_lbl: find result in a list; sum findings, mandatory
    :params: use_file_system: True 'use_file_system'; False: in mem unzip
    :params: url: dl_url, if not 'use_file_system', URL of csv file
    :params: zipped_filename: 'revised.csv',  name of the uncompressed file in zip archive
    :params: csv_col_name: 'value',  # table column header name, mandatory
    :params: 'num_rows': num list rows to calc in one loop for each process
    :params: 'num_proc': procs to start, can be more or less than system CPU core count
    """
    mP = eisenmp.Mp()
    mP.start(**modConf.__dict__)  # thread & processes start, attributes avail. for worker and feeder loop
    mP.mp_print_q.put('\tDownload large list')
    mP.mp_print_q.put(modConf.dl_url)

    report_file = os.path.join(dir_name, 'download', 'report.zip')
    downloader = download_report_zip_archive(mP, report_file)  # and prn msg
    generator = g_use_fs_csv(report_file) if modConf.use_file_system else g_in_mem_csv(downloader)  # CSV generator
    mP.run_q_feeder(generator=generator)

    return mP


def g_in_mem_csv(downloader):
    """
    """
    archive = downloader.unzip_mem()
    in_file = archive.open(modConf.zipped_filename, 'r')
    dict_reader = csv.DictReader(TextIOWrapper(in_file, 'utf-8'))
    generator = (column[modConf.csv_col_name] for column in dict_reader)
    return generator


def g_use_fs_csv(report_file):
    """
    """
    file_path = os.path.join(os.path.dirname(report_file), modConf.zipped_filename)
    filename = open(file_path, 'r')
    dict_reader = csv.DictReader(filename)
    
    generator = (column[modConf.csv_col_name] for column in dict_reader)
    return generator


def unzip_on_file_system(downloader, report_f):
    """
    """
    downloader.save(report_f)
    with ZipFile(report_f, 'r') as un_zip:
        un_zip.extractall(os.path.dirname(report_f))


def download_report_zip_archive(mP, report_file):
    """
    """
    mP.mp_print_q.put(origin_msg_create())
    downloader = DownLoad()
    downloader.url = modConf.url
    downloader.zipped_filename = modConf.zipped_filename
    downloader.load_url()  # now decide to read csv from mem or file
    if modConf.use_file_system:
        unzip_on_file_system(downloader, report_file)  # comment no dl
    return downloader


def origin_msg_create():
    """
    """
    str_fs = 'WRITE_FILE_SYSTEM'
    str_dl = 'LOAD_URL_UNZIP_READ_IN_MEM'
    origin = str_fs if modConf.use_file_system else str_dl
    origin_msg = f'\n\t{origin} [{modConf.zipped_filename}] [{modConf.csv_col_name}]' \
                 f'\n\t~~~~~~~~~~~~~~~~\n'
    return origin_msg


def main():
    """
    """
    start = time.perf_counter()

    mP = generator_calc_csv()
    while 1:
        # running threads, wait
        if mP.begin_proc_shutdown:
            break
        time.sleep(1)

    print(f'Time in sec: {round((time.perf_counter() - start))}')

    msg_result = e_utils.Result.result_dict
    return msg_result


if __name__ == '__main__':
    main()



import os
from Flask_SQLAlchemy_Project_Template import create_app, setup_database, db_path


def worker(toolbox):  # name this arg as you like
    """
    - Worker -  eisenmp_exa_wrk_flask_orm_srv_one_cpu

    toolbox is the all-in-one box for vars and queues. incl. ModuleConfiguration
    """
    color_dict = {
        'PURPLE': '\033[1;35;48m',
        'CYAN': '\033[1;36;48m',
        'BOLD': '\033[1;37;48m',
        'BLUE': '\033[1;34;48m',
        'GREEN': '\033[1;32;48m',
        'YELLOW': '\033[1;33;48m',
        'yellow': '\033[1;31;48m',
        'BLACK': '\033[1;30;48m',
        'UNDERLINE': '\033[4;37;48m',
        'END': '\033[1;37;0m',
    }

    port, col = 0, None
    if not toolbox.WORKER_ID % 2:   # mod is 1 odd
        col = color_dict['BLUE']
        port = blue_q_get(toolbox)[1]  # [0] is header row
    if toolbox.WORKER_ID % 2:
        col = color_dict['YELLOW']
        port = yellow_q_get(toolbox)[1]

    col_end = color_dict['END']

    msg = col + f'\nWORKER_MSG worker: {toolbox.WORKER_ID} pid: {toolbox.WORKER_PID} server port: {port}' + col_end
    toolbox.mp_print_q.put(msg)

    # Flask
    app_factory = create_app(port)  # flask, we feed port number to update the route -> Html page with our address
    if not os.path.isfile(db_path):  # do not kill db, if exists; MUST exist if many srv, else create by many srv, crash
        setup_database(app_factory)
    app_factory.run(host="localhost", port=port)


def blue_q_get(toolbox):
    """"""
    while 1:
        if not toolbox.mp_blue_q.empty():
            port_lst = toolbox.mp_blue_q.get()  # has header with serial number
            return port_lst


def yellow_q_get(toolbox):
    """"""
    while 1:
        if not toolbox.mp_yellow_q.empty():
            port_lst = toolbox.mp_yellow_q.get()  # has header with serial number
            return port_lst





# Python 3 server example
# https://pythonbasics.org/webserver/
# mod by 44xtc44
import os

import json
import time
from http.server import BaseHTTPRequestHandler, HTTPServer

try:
    import eisenmp
    import eisenmp.eisenmp_exa_multi_srv_each_cpu as multi_on_each_cpu
    import eisenmp.eisenmp_exa_each_flask_orm_srv_one_cpu as single_on_each_cpu
    import eisenmp.eisenmp_exa_prime as prime
    import eisenmp.eisenmp_exa_web_csv as web_csv
    import eisenmp.eisenmp_exa_http as http_srv
    import eisenmp.eisenmp_exa_double_q as double_q
    import eisenmp.eisenmp_exa_bruteforce as bruteforce

except ImportError:
    import eisenmp_examples
    import eisenmp_examples.eisenmp_exa_multi_srv_each_cpu as multi_on_each_cpu
    import eisenmp_examples.eisenmp_exa_each_flask_orm_srv_one_cpu as single_on_each_cpu
    import eisenmp_examples.eisenmp_exa_prime as prime
    import eisenmp_examples.eisenmp_exa_web_csv as web_csv
    import eisenmp_examples.eisenmp_exa_http as http_srv
    import eisenmp_examples.eisenmp_exa_double_q as double_q
    import eisenmp_examples.eisenmp_exa_bruteforce as bruteforce

hostName = "localhost"
serverPort = 12_321
dir_name = os.path.dirname(__file__)


class Menu:
    example_menu = [
        'Multiple server in each process/CPU core - share a port range - share a DB. Worker Green, Yellow and Blue',
        'Prime Number calculation',
        'Every Flask server in a different process - share a port range - share a DB. Worker Yellow and Blue',
        'Web CSV large list. Average for each chunk of a large list column to simply calc the results later.',
        'One simple http server presents a radio on every process. "Unemployed" worker exit.',
        'Each process has two Queues feed audio and video. Merge in a fake production.',
        'Brute force attack with dictionary and itertools generator'
    ]

    exa_tpl_lst = [
        (0, example_menu[0], multi_on_each_cpu.main),
        (1, example_menu[1], prime.main),
        (2, example_menu[2], single_on_each_cpu.main),
        (3, example_menu[3], web_csv.main),
        (4, example_menu[4], http_srv.main),
        (5, example_menu[5], double_q.main),
        (6, example_menu[6], bruteforce.main)
    ]


def run_http():
    """Blocked, no loop here
    """
    global serverPort

    while 1:
        try:
            webServer = HTTPServer((hostName, serverPort), MyServer)
            break
        except OSError:
            print(f'serverPort in use {serverPort}\n\tadd one to port number')
            serverPort += 1  # port already in use

    print(f"Examples http://%s:%s" % (hostName, serverPort))
    try:
        webServer.serve_forever()
    except KeyboardInterrupt:
        pass
    webServer.server_close()
    print("Server stopped.")


class MyServer(BaseHTTPRequestHandler):

    def do_POST(self):
        """"""
        # print(self.headers)
        length = int(self.headers.get_all('content-length')[0])
        print(self.headers.get_all('content-length'))
        data_string = self.rfile.read(length)
        example_fun = Menu.exa_tpl_lst[int(data_string)][2]
        ret_val = example_fun()
        # print(data_string)
        self.send_response(200)
        # self.send_header("Content-type", "text/plain")
        self.send_header('Content-Type', 'application/json')
        self.end_headers()
        self.flush_headers()
        json_string = json.dumps(str(ret_val))
        self.wfile.write(bytes(json_string, "utf-8"))
        return True

    def do_GET(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()

        da_html_lst = [

            "<!doctype html><html><html lang='en'>",
            "<head>",
            "<meta charset='UTF-8'>",
            "<meta X-Content-Type-Options='nosniff'>",
            "<meta name='viewport' content='width=device-width, initial-scale=1.0'>",
            "<style>",
            "@font-face {font-family: 'PT Sans';font-style: normal;font-weight: normal;",
            "src: local('PT Sans'), local('PTSans-Regular'),",
            "url(data:application/font-woff2;charset=utf-8;base64,d09GRgABAAAAAHowABMAAAAA+OAA) format('woff2');}",

            "body{font-family:PT Sans, arial; color:black;background-color:#ccc;}",
            "h1 {color:brown;text-align:center;}",
            ".container {display: flex;}",
            ".middle {display: flex;flex-direction: column;margin: auto;color:brown;}",
            ".pMid {color: black; text-align:center;}",
            "</style>",
            f"<title>Examples</title>",
            "</head>",

            "<body>",
            f"<h1> Welcome to eisenmp_examples</h1>",
            f"<p class=pMid> Examples run in the terminal window. </p>",
            f"<div class='container'>",
            f"<div class='middle'>",
            '_o__example_buttons____',
            "</div></div>",
            "send: <p id=pRspReturn class=pInOut> </p>",
            f"resp: <p id=pMsg class=pInOut> </p>",
            "</body></html>",

            "<script>",
            "function getExa(radio_btn_id) {",
            "document.getElementById('pMsg').innerHTML='';",
            "const xhr = new XMLHttpRequest();",
            "xhr.open('POST', '/');",
            "xhr.onload = function () {console.log('xhr r ', xhr.response); ",
            "document.getElementById('pMsg').innerHTML=xhr.response;}; ",
            "xhr.send(radio_btn_id);",
            "let pRspReturn = document.getElementById('pRspReturn');",
            "pRspReturn.innerText=radio_btn_id;",
            "}",
            "</script>",
            "</body></html>"
        ]

        for _ in da_html_lst:
            if _ == '_o__example_buttons____':
                for i, _ in enumerate(Menu.exa_tpl_lst):
                    idx = Menu.exa_tpl_lst[i][0]
                    show = Menu.exa_tpl_lst[i][1]
                    exa_html_line = f"<div><label><input type='radio' name='da'" \
                                    f"id='{idx}' onclick='getExa(id)'>{show}</label></div>"
                    self.wfile.write(bytes(exa_html_line, "utf-8"))
                continue
            self.wfile.write(bytes(_, "utf-8"))


def main():
    """
    """
    start = time.perf_counter()

    run_http()

    print('Time in sec: ', round((time.perf_counter() - start)))


if __name__ == '__main__':
    main()



import io
import zipfile
from io import BytesIO

try:
    import eisenmp.utils_exa.eisenmp_utils as e_utils
except ImportError:
    import eisenmp_examples.utils_exa.eisenmp_utils as e_utils


class DownLoad:
    """If we have the response, we can read it at once or as a stream.

    """

    def __init__(self):
        self.url = None
        self.zipped_filename = None  # one file in zip container
        self.csv_col = None
        self.response = None
        self.file_bin = None

    def load_url(self):
        """
        """
        self.response = e_utils.load_url(self.url)

    def save(self, file_name):
        """Save response on file system with OS buf size.
        Dotted output in one line.
        """
        with open(file_name, 'wb') as file_writer:
            i = 0
            while 1:
                if not (i % 100):
                    print('.', end="", flush=True)
                chunk = self.response.read(io.DEFAULT_BUFFER_SIZE)
                if not chunk:
                    print('\n')
                    break
                file_writer.write(chunk)
                i += 1

    def unzip_mem(self):
        """Extract zip in memory to avoid unwanted files
        from archive on file system.
        """
        zip_file_mem = BytesIO(self.response.read())
        readable_archive = zipfile.ZipFile(zip_file_mem)
        return readable_archive

    def unzip_fs(self):
        """"""
        pass

    def store_mem(self):
        self.file_bin = bytearray()
        i = 0
        while 1:
            if not (i % 100):
                print('.', end="", flush=True)
            chunk = self.response.read(io.DEFAULT_BUFFER_SIZE)
            if not chunk:
                print('\n')
                break
            self.file_bin += chunk
            i += 1


if __name__ == '__main__':
    pass




import os
import ssl
import time
import random
import urllib
import certifi
import threading
from hashlib import sha256
from zipfile import ZipFile
from urllib.request import urlopen, Request
from collections import defaultdict


os.environ['SSL_CERT_FILE'] = certifi.where()
context_ssl = ssl.create_default_context(cafile=certifi.where())

agent_list = [   # agent orange or agent white
    'Mozilla/5.0 (iPad; CPU OS 8_4_1 like Mac OS X) AppleWebKit/600.1.4 (KHTML, like Gecko) '
    'Version/8.0 Mobile/12H321 Safari/600.1.4',
    'Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) '
    'Chrome/45.0.2454.85 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) '
    'Chrome/45.0.2454.85 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) '
    'Chrome/42.0.2311.135 Safari/537.36 Edge/12.10240',
    'Mozilla/5.0 (Windows NT 6.3; WOW64; rv:40.0) Gecko/20100101 Firefox/40.0',
]


class Result:
    """Finest results only here.
    Can save number stuff, not response chunks, or we crash

    """

    def __init__(self):
        self._result_dict = defaultdict(list)  # threads can read/write dict

    @property
    def result_dict(self):
        return self._result_dict

    @result_dict.setter
    def result_dict(self, update):
        self._result_dict = update

    @result_dict.deleter
    def result_dict(self):
        del self._result_dict

    def result_dict_update(self, key, value):
        self._result_dict[key].append(value)


def consecutive_number():
    """Want a stamp on each list header.
    Used for Queue messages get() and put in box_dict[num] = msg
    Can rebuild original order if worker puts result in a list
    with same num as order list.
    """
    idx = 0
    while 1:
        yield idx
        idx += 1


def load_url(url, user_agent=None):
    """Get server response.

    :params: url: url
    :params: user_agent: override python ua of urllib (we are a browser)
    :exception: Timeout recursive call
    :return: http server response
    :rtype: http response
    """
    if not user_agent:
        user_agent = agent_list[random.randint(0, len(agent_list) - 1)]
    try:
        request = Request(url)
        opener = urllib.request.build_opener()
        if user_agent:
            opener.addheaders = [('User-agent', user_agent)]
        urllib.request.install_opener(opener)
        response = urlopen(request, timeout=15, context=context_ssl)
        return response

    except TimeoutError:
        print(f'TimeoutError in load_url().')
        return False
    except Exception as error:
        print(f'unknown error in load_url() {error}')
        return False


def thread_shutdown_wait(*threads):
    """We return if none of the thread names are listed anymore.
    Blocks!

    :params: *threads: arbitrary list of thread names
    """
    busy = True
    while busy:
        names_list = [thread.name for thread in threading.enumerate()]
        busy = True if any([True for thread in threads if thread in names_list]) else False
        time.sleep(.1)


def condense_list_from_fs(search_str, words_dict):
    """
    """
    short_list = [word for word in words_dict.keys() if str_with_len_get(word, search_str)]
    return short_list


def str_with_len_get(word, str_to_comp):
    if len(word) == len(str_to_comp):

        return word
    return False


def split_list(lst, chunk_size):
    """Needs a companion loop.
    for chunk in split_list(word_list, chunk_size): print(chunk)

    :params: lst: the list to split
    :params: chunk_size: row count of lst for one consumer
    """
    for i in range(0, len(lst), int(chunk_size)):
        yield lst[i: i + int(chunk_size)]


def merge_list(*list_paths, lowercase=True):
    """WORD LISTS.
    We create a dict with word as key or hash digest, for speed.

    :params: args: tuple of OS paths word list(s)
    :params: lower: True is all lowercase
    """
    word_list = []
    for arg in list(*list_paths):
        print(f' .. read wordlist {arg}')
        with open(arg, 'r', encoding="UTF-8") as reader:
            if lowercase:
                stripped = [line.lower().rstrip() for line in reader.readlines()]  # remove \n
                pass
            else:
                stripped = [line.rstrip() for line in reader.readlines()]
            word_list.extend(stripped)
    return word_list


def unzip(file_path):
    """Extractor.

    :params: file_path: to zipped file
    """
    with ZipFile(file_path, 'r') as un_zip:
        un_zip.extractall(os.path.dirname(file_path))


def create_hash(word):
    """Inputs a string and returns the sha256 digest

    :params: word: to hash
    """
    w_bytes = word.encode('utf-8')
    w_hash = sha256(w_bytes)
    digest = w_hash.hexdigest()
    return digest


def replace_special_char(word):
    """German so far.

    :params: word: string to check
    """
    if 'ã¤' in word:
        word = word.replace('ã¤', 'ae')
    elif 'ã¼' in word:
        word = word.replace('ã¼', 'ue')
    elif 'ä¶' in word:
        word = word.replace('ä¶', 'oe')
    elif 'ã¶' in word:
        word = word.replace('ã¶', 'oe')
    elif 'ãÿ' in word:
        word = word.replace('ãÿ', 'ss')
    elif 'Ã„' in word:
        word = word.replace('Ã„', 'AE')
    elif 'Ãœ' in word:
        word = word.replace('Ãœ', 'UE')
    elif 'Ã–' in word:
        word = word.replace('Ã–', 'OE')
    elif 'ä' in word:
        word = word.replace('ä', 'ae')
    elif 'Ä' in word:
        word = word.replace('Ä', 'AE')
    elif 'ü' in word:
        word = word.replace('ü', 'ue')
    elif 'Ü' in word:
        word = word.replace('Ü', 'UE')
    elif 'ö' in word:
        word = word.replace('ö', 'oe')
    elif 'Ö' in word:
        word = word.replace('Ö', 'OE')
    elif "'" in word:
        word = word.replace("'", '')
    else:
        pass
    return word



import itertools

try:
    import eisenmp.utils_exa.eisenmp_utils as e_utils
except ImportError:
    import eisenmp_examples.utils_exa.eisenmp_utils as e_utils


class SearchStr:
    """Hosts the dict to compare,
    and a 'search string shaker' generator.

    el cheapo goes brute force, but it shows the basics
    go further with 'create_key_digest_val_word_dict' and load
    later a rainbow table with the 'mp_tools_q' into each process

    """

    def __init__(self):
        super().__init__()
        self.search_string = ''
        self.loader_words_dict = {}  # THE dict
        self.words_dict = {}  # work with this dict
        self.csv_column = []  # free to use list, should have one row
        self.search_a_str_thread_name = 'eisenmp_search_a_str_thread'

    def create_key_word_val_none_shrink(self, lowercase=True):
        """Optimized 'create_key_word_val_none_dict'.
        Half of the time. Shrink dict to match only search str length.

        :params: args: OS paths to word lists
        :params: lower: can convert to lower string
        """
        shrink_list = e_utils.condense_list_from_fs(self.search_string, self.loader_words_dict)

        if lowercase:
            self.words_dict = {word.lower(): None for word in shrink_list}
            return

        self.words_dict = {word: None for word in shrink_list}  # None value

    def create_key_word_val_none_dict(self, *args, lowercase=True):
        """{'aal': None, ...}
        Create a dict from wordlists (procs mem inflates to 10 x word_list_size in mb). Faster reading vs list.
        - Dict. Each process gets a copy of the dict vs Multiprocess shared manager dict, which is utter slow.
        - Merged List. Add as many as you like.

        :params: args: OS paths to word lists
        :params: lower: can convert to lower string
        """
        if lowercase:
            self.loader_words_dict = {word: None for word in e_utils.merge_list(*args, lowercase=lowercase)}
            return self.words_dict

        self.loader_words_dict = {word: None for word in e_utils.merge_list(*args)}  # None value

    def create_key_digest_val_word_dict(self, *args, lower=None):
        """{'83c54220e5f2c521819cb6d80163858dd6def3c5a9ed37281a532284b342104a': 'aal',}
        Creates a ``real`` sha256 hash dict for pwd bruteforce.

        :params: args: OS paths to word lists
        :params: lower: can convert to lower string
        """
        if lower:
            wd = {digest: word.lower() for digest, word in
                  zip(map(e_utils.create_hash, e_utils.merge_list(*args)),
                      e_utils.merge_list(*args))}
            self.loader_words_dict = wd  # key is hash output from list (must be lower()): value is string lower()
            return

        self.loader_words_dict = {digest: word for digest, word in
                                  zip(map(e_utils.create_hash, e_utils.merge_list(*args)),
                                      e_utils.merge_list(*args))}

    def generator(self, lowercase=True):
        """STRING generator with all permutations of init_str.
        Itertools creates duplicates of a string permutation.
        We could redirect output to dict, key val, but 15! is a Terabyte string thingy. Needs custom, paid ;), solution.
        - Permutation list generator. 15! is 1,307,674,368,000; 'AEEFFGILNNOPRTT' not solved yet
        1.3 Trillion x 10 byte (guess for a list); 12,770,257,500 kb is 12,7702575 Terabyte list size
        A 16 core XEON will need approximately 20 days for a 15! string, to generate and test all permutations.
        Looks like linear Ordnung to calculate for other core counts. See 'ProcInfo' how to calc ETA.

        :params: init_str: string to shuffle and shake 'DIKKLOOR' -> results in german KROKODIL
        """
        s_str = self.search_string if not lowercase else self.search_string.lower()

        for permutation in itertools.permutations(s_str):
            yield ''.join(permutation)


if __name__ == '__main__':
    ss = SearchStr()



"""Example 1 brute force

eisenmp uses a list reducer module in example 2.
Generated list rows from language dict are compared with chars.

Vars and Queues stored in the toolbox instance.
See documentation for a quick overview, please.

"""


def worker_entrance(toolbox):
    """
    - WORKER - Called in a loop.

    Start, Entry, Exit of this single process worker.
    We return True to get next list chunk, whatever object is in the rows.
    Fed from mp_input_q to our toolbox. toolbox is our work instance with queues,
    messages, list chunk, and work tools like language dictionary or hash list.

    toolbox.foo, gives also access to all attributes and values
    of the 'modConf.foo' instance, you have created
    """
    if toolbox.multi_tool_get:
        tool_get(toolbox)  # dict mp_tools_q
    workload_get(toolbox)
    busy = brute_force(toolbox)  # worker function
    if not busy:
        return False
    send_eta_data(toolbox)
    return True


def tool_get(toolbox):
    """"""
    while 1:
        if not toolbox.mp_tools_q.empty():
            toolbox.multi_tool = toolbox.mp_tools_q.get()
            toolbox.multi_tool_get = False  # can be set in modConf
            break


def workload_get(toolbox):
    """"""
    while 1:
        if not toolbox.mp_input_q.empty():
            toolbox.NEXT_LIST = toolbox.mp_input_q.get()
            toolbox.num_lists += 1
            break


def remove_header(toolbox):
    """Transport ticket with consecutive number.
    Remove if no recreation of order is necessary.
    Can reuse list for result, if rebuild order.

    Use self.header_msg attribute to overwrite default header string
    """
    # toolbox.mp_print_q.put(toolbox.NEXT_LIST[0])
    toolbox.INPUT_HEADER = toolbox.NEXT_LIST[0]
    del toolbox.NEXT_LIST[0]  # remove header str


def send_output(toolbox, formatted_str):
    """Put your findings in the output list.
    Find results in the 'eisenmp_utils.Result.result_dict'

    :params: toolbox: -
    :params: average: average of the (chunk of) column
    """
    # header for output result list
    header = toolbox.OUTPUT_HEADER + toolbox.INPUT_HEADER  # q collector can distinguish queues and store result in dict
    result_lst = [header,
                  formatted_str]  # your findings here
    toolbox.mp_output_q.put(result_lst)


def send_eta_data(toolbox):
    """list of [PERF_HEADER_ETA, PERF_CURRENT_ETA] to ProcInfo, to calc arrival time ETA
    """
    toolbox.PERF_CURRENT_ETA = len(toolbox.NEXT_LIST)
    perf_lst = [toolbox.PERF_HEADER_ETA + toolbox.WORKER_NAME,  # binary head
                toolbox.PERF_CURRENT_ETA]
    toolbox.mp_info_q.put(perf_lst)  # ProcInfo calc arrival time and % from mp_info_q, of all proc lists


def brute_force(toolbox):
    """List -> dict str compare until stop order in last list, generator empty.

    :params: worker_msg: test for a valid string
    """
    busy = True
    if toolbox.STOP_MSG in toolbox.NEXT_LIST:  # eisenmp.iterator_loop() informs stop, no more lists
        busy = False  # loop worker sends shutdown msg to next worker - generator is empty
    remove_header(toolbox)  # remove if no reassembling

    for str_permutation in toolbox.NEXT_LIST:  # 'iiattbz' string permutation in the row
        search_str(str_permutation, toolbox)
    return busy


def search_str(s_str, toolbox):
    """Write to result Q if str matched.

    :params: s_str: current generated string to test
    :params: multi_tool: here the words dict
    """
    if s_str in toolbox.multi_tool:  # match dict
        send_output(toolbox, f'... proc {toolbox.WORKER_NAME} ... {s_str}')

        toolbox.mp_print_q.put(f'... proc {toolbox.WORKER_NAME} ... {s_str}')



"""

"""
import time


def worker_entrance(toolbox):
    """
    - WORKER - Called in a loop.
    """
    audio_chunk_lst, video_chunk_lst = None, None
    if not toolbox.WORKER_ID % 2:  # mod is 1 odd
        audio_chunk_lst = batch_1_audio_get(toolbox)
        video_chunk_lst = batch_1_video_get(toolbox)  # batch_1_video_get(toolbox) ['head_1', 'foo', 'bar', 'buz']
    if toolbox.WORKER_ID % 2:  # mod is 0 even
        audio_chunk_lst = batch_7_audio_get(toolbox)
        video_chunk_lst = batch_7_video_get(toolbox)
    print(f'....{toolbox.WORKER_ID} {audio_chunk_lst} {video_chunk_lst}')
    busy = template_worker(toolbox, audio_chunk_lst, video_chunk_lst)  # worker function
    if not busy:
        return False
    return True


def batch_1_video_get(toolbox):
    """"""
    while 1:
        if not toolbox.batch_1['video_in'].empty():
            lst = toolbox.batch_1['video_in'].get()
            toolbox.num_of_lists += 1  # list counter prn screen
            return lst


def batch_1_audio_get(toolbox):
    """"""
    while 1:
        if not toolbox.batch_1['audio_lg'].empty():
            lst = toolbox.batch_1['audio_lg'].get()
            return lst


def batch_7_video_get(toolbox):
    """"""
    while 1:
        if not toolbox.batch_7['video_in'].empty():
            lst = toolbox.batch_7['video_in'].get()
            toolbox.num_of_lists += 1  # list counter prn screen
            return lst


def batch_7_audio_get(toolbox):
    """"""
    while 1:
        if not toolbox.batch_7['audio_lg'].empty():
            lst = toolbox.batch_7['audio_lg'].get()
            return lst


def remove_header(lst):
    """Transport ticket with consecutive number.
    Remove if no recreation of order is necessary.
    Can reuse list for result, if rebuild order.

    Use self.header_msg attribute to overwrite default header string
    """
    del lst[0]  # remove header str


def send_eta_data(toolbox, lst):
    """list of [PERF_HEADER_ETA, PERF_CURRENT_ETA] to ProcInfo, to calc arrival time ETA
    """
    toolbox.PERF_CURRENT_ETA = len(lst)
    perf_lst = [toolbox.PERF_HEADER_ETA + toolbox.WORKER_NAME,  # binary head
                toolbox.PERF_CURRENT_ETA]
    # disable info q will block all
    toolbox.mp_info_q.put(perf_lst)  # ProcInfo calc arrival time and % from info_q, of all proc lists


def send_output(toolbox, row_aud, row_vid):
    """Put your findings in the output list.
    Find results in the 'eisenmp_utils.Result.result_dict'

    :params: toolbox: -
    :params: res_lst: list, res_lst = [row_aud, row_aud]
    """
    # header for output result list
    header = toolbox.OUTPUT_HEADER + toolbox.header_aud  # q collector can distinguish qs and store result in dict
    result_lst = [header,
                  row_aud]  # your findings here
    toolbox.mp_output_q.put(result_lst)

    header = toolbox.OUTPUT_HEADER + toolbox.header_vid  # q collector can distinguish qs and store result in dict
    result_lst = [header,
                  row_vid]  # your findings here
    toolbox.mp_output_q.put(result_lst)


def template_worker(toolbox, audio_chunk_lst, video_chunk_lst):
    """
    """
    busy = True
    toolbox.header_aud = audio_chunk_lst[0]
    toolbox.header_vid = video_chunk_lst[0]

    remove_header(audio_chunk_lst)  # remove list header with serial number if no reassembling
    remove_header(video_chunk_lst)

    for idx, row_aud in enumerate(audio_chunk_lst):
        row_aud = row_aud
        row_vid = video_chunk_lst[idx]
        pass
        if toolbox.STOP_MSG in str(row_aud) or toolbox.STOP_MSG in str(row_vid):  # stop is str
            return False
        else:
            msg = f'worker: {toolbox.WORKER_ID} cat: {toolbox.header_aud} {toolbox.header_vid}' \
                  f'audio: {row_aud} vid: {row_vid} list({toolbox.num_of_lists})'
            msg_col = None
            if not toolbox.WORKER_ID % 2:
                msg_col = toolbox.BLUE
            if toolbox.WORKER_ID % 2:
                msg_col = toolbox.RED

            msg_show = msg_col + msg + toolbox.END
            toolbox.mp_print_q.put(msg_show)
            # output result
            send_output(toolbox, row_aud, row_vid)
            time.sleep(.2)

        send_eta_data(toolbox, audio_chunk_lst)
    return busy



"""See some examples,
it's always the same process,
you extend or switch off defaults

"""
import os
import time
import platform
import eisenmp


class ModuleConfiguration:
    """
    You can use the class to have your variables available in the module.

    """
    dir_name = os.path.dirname(__file__)  # our module path without file name
    # path to worker module and entry function reference, worker module import in [isolated] process environment
    # -------------------- MANDATORY WORKER STRINGS --------------------
    first_module = {
        'WORKER_PATH': os.path.join(dir_name, 'worker', 'eisenmp_exa_wrk_flask_orm_srv_one_cpu.py'),
        'WORKER_REF': 'worker',  # Warning: loader runs all f() with a single argument 'toolbox'; pull args from it
    }
    foo = {'WORKER_PATH': 'bar', 'WORKER_REF': 'baz'}

    def __init__(self):
        # load order list, first module is called in an endless loop, you can append your own loop inside the worker
        self.worker_modules = [self.first_module]

        # Multiprocess vars - override default
        self.NUM_PROCS = 6  # your process count, default is None: one proc/CPU core
        self.NUM_ROWS = 1  # tell iterator to make only one list row, each worker needs only one number
        self.RESULTS_STORE = True  # keep in dictionary, will crash the system if store GB network chunks in mem
        self.RESULTS_PRINT = True  # result rows of output are collected in a list, display if processes are stopped
        self.RESULT_LABEL = 'No result, server blocks'  # pretty print as result header for RESULTS_PRINT
        self.RESULTS_DICT_PRINT = False  # shows content of results dict with ticket numbers, check tickets
        if platform.system() == 'Linux':
            self.START_METHOD = 'fork'  # 'spawn' is default if unused; also use 'forkserver' or 'fork' on Unix only


modConf = ModuleConfiguration()  # Accessible in the module.


def manager():
    """
    - Manager -

    !!! Database must be [created with one process only], then many procs can read, write !!!
    """
    # need a Queue for yellow and blue and an `existing` Database with numbers or generator range step
    q_name_maxsize = [
        # q_name, q_maxsize;
        ('mp_blue_q', 1),  # tuple, worker: toolbox.mp_blue_q.get()
        ('mp_yellow_q', 1)
    ]
    # default call
    mP = eisenmp.Mp()

    # custom queues for port groups ---> need a generator for each queue
    mP.queue_cust_dict_std_create(*q_name_maxsize)  # unpack, create Qs in std {default} dict ..['mp_blue_q']=Queue()

    port_blue = (port_number for port_number in range(13_000, 13_006, 2))
    port_yellow = (port_number for port_number in range(14_000, 14_006, 2))

    # !!! config write instance dictionary if all args set !!!
    mP.start(**modConf.__dict__)  # feed toolbox, instance attributes available for worker and feeder loop
    mP.run_q_feeder(generator=port_blue, input_q=mP.queue_cust_dict_std['mp_blue_q'])
    mP.run_q_feeder(generator=port_yellow, input_q=mP.queue_cust_dict_std['mp_yellow_q'])


def main():
    """
    """
    start = time.perf_counter()

    manager()
    msg = f'\nFlask ORM - each one proc, Time in sec: {round((time.perf_counter() - start))} - main() exit'
    print(msg)
    return msg


if __name__ == '__main__':
    main()



"""See some examples,
it's always the same process,
you extend or switch off defaults


Multiple server on each CPU core, if server has that many cores.
needs Flask_SQLAlchemy_Project_Template >=1.3
"""
import os
import time
import platform
import eisenmp


class ModuleConfiguration:
    """
    You can use the class to have your variables available in the module.

    """
    dir_name = os.path.dirname(__file__)  # our module path without file name
    # path to worker module and entry function reference, worker module import in process environment
    # -------------------- MANDATORY WORKER STRINGS --------------------
    first_module = {
        'WORKER_PATH': os.path.join(dir_name, 'worker', 'eisenmp_exa_wrk_multi_srv_each_cpu.py'),
        'WORKER_REF': 'worker',  # Note: loader calls all f() with a single argument 'toolbox'; worker(toolbox)
    }
    foo = {'WORKER_PATH': 'bar', 'WORKER_REF': 'baz'}

    def __init__(self):
        # load order list, first module is called in an endless loop if it is not a thread
        self.worker_modules = [  # multi server on one CPU core
            self.first_module,
            self.first_module,
            self.first_module,  # each module_loader in the process loads this list
            self.first_module,
            self.first_module,
            self.first_module,  # just demo, same can be achieved if server called in a worker loop
        ]

        # Multiprocess vars - override default
        self.NUM_PROCS = 3  # your process count, default is None: one proc/CPU core
        self.NUM_ROWS = 1  # tell iterator to make only one list row, each worker needs only one number
        self.RESULTS_STORE = True  # keep in dictionary, will crash the system if store GB network chunks in mem
        self.RESULTS_PRINT = True  # result rows of output are collected in a list, display if processes are stopped
        self.RESULTS_DICT_PRINT = False  # shows content of results dict with ticket numbers, check tickets
        if platform.system() == 'Linux':
            self.START_METHOD = 'fork'  # 'spawn' is default if unused; also use 'forkserver' or 'fork' on Unix only

        self.STOP_MSG_DISABLE = True  # module_loader leaves worker loop and waits for stop msg in mp_process_q

        # worker port groups, nailed on one cpu
        self.blue_lst = [0]  # one CPU core for blue list, if toolbox.kwargs['START_SEQUENCE_NUM'] in worker_blue_lst,
        self.yellow_lst = [1]  # one CPU core for yellow, process spawn num 1; (class ProcEnv 'run_proc -> core')
        self.green_lst = [2]


modConf = ModuleConfiguration()  # Accessible in the module.


def manager():
    """
    - Manager -

    !!! Database must be [created with one process only], then many procs can read, write !!!


    ORM https://en.wikipedia.org/wiki/Object%E2%80%93relational_mapping
    """
    # need a Queue for yellow and blue and an `existing` Database with numbers or generator range step
    q_name_maxsize = [
        # q_name, q_maxsize;
        ('mp_blue_q', 1),  # tuple, worker: toolbox.mp_blue_q.get()
        ('mp_yellow_q', 1),
        ('mp_green_q', 1)
    ]
    # default call
    mP = eisenmp.Mp()

    # custom queues for port groups ---> need a generator for each queue
    mP.queue_cust_dict_std_create(*q_name_maxsize)  # unpack, create Qs in std {default} dict ..['mp_blue_q']=Queue()

    # !!! config write instance dictionary if all args set !!!
    mP.start(**modConf.__dict__)  # feed toolbox, instance attributes available for worker and feeder loop

    port_generator_blue = (port_number for port_number in range(11_000, 11_006, 1))
    port_generator_yellow = (port_number for port_number in range(14_000, 14_006, 1))
    port_generator_green = (port_number for port_number in range(15_000, 15_006, 1))
    mP.run_q_feeder(generator=port_generator_blue, input_q=mP.queue_cust_dict_std['mp_blue_q'])
    mP.run_q_feeder(generator=port_generator_yellow, input_q=mP.queue_cust_dict_std['mp_yellow_q'])
    mP.run_q_feeder(generator=port_generator_green, input_q=mP.queue_cust_dict_std['mp_green_q'])


def main():
    """
    """
    start = time.perf_counter()

    manager()

    msg_time = f'\nMulti loader Time in sec: {round((time.perf_counter() - start))} - main() exits'
    print(msg_time)
    return msg_time


if __name__ == '__main__':
    main()



"""Multiprocessor Prime number finder (worker) and number Generator.

Example for a ``two in one module``. Manager and worker.

This is NOT ADVISABLE and only to show the mechanics.
Use a separate worker module to not import your 'Generator' imports twice.
That is what the loader is for.
"""
import os
import time
import math
import platform

import eisenmp
import eisenmp.utils.eisenmp_utils as e_utils

dir_name = os.path.dirname(__file__)


class ModuleConfiguration:
    """
    You can use the class to have your variables available in the module.

    'worker_prime()' is executed by all processes on a CPU somewhere.
    Each 'worker_prime()' gets (one by one) a list chunk from eisenmp.mp_input_q.
    You have full access to all queues and methods. mp = eisenmp.Mp()
    """
    # path to worker module and entry function reference, worker module import in [isolated] process environment
    # -------------------- MANDATORY WORKER STRINGS --------------------
    first_module = {
        'WORKER_PATH': os.path.join(dir_name, 'eisenmp_exa_prime.py'),
        'WORKER_REF': 'worker_prime',  # loader runs worker_prime(toolbox) to provide all args from toolbox (kwargs)
    }
    foo = {'WORKER_PATH': 'bar', 'WORKER_REF': 'baz'}

    def __init__(self):
        # load order list, first module is called in an endless loop
        self.worker_modules = [
            self.first_module,   # second module must be threaded, else we hang
            # foo
        ]

        # Multiprocess vars - override default
        self.NUM_PROCS = 5  # your process count, default is None: one proc/CPU core
        # max generator / NUM_ROWS = number of tickets, 10_000 / 42 = 238.095 -> 238 lists with ticket numbers
        self.NUM_ROWS = 42  # your workload spread, list (generator items) to calc in one loop, default is None: 1_000
        self.RESULTS_STORE = True  # keep in dictionary, will crash the system if store GB network chunks in mem
        self.RESULTS_PRINT = True  # result rows of output are collected in a list, display if processes are stopped
        self.RESULTS_DICT_PRINT = True  # shows content of results dict with ticket numbers, check tickets
        if platform.system() == 'Linux':
            self.START_METHOD = 'fork'  # 'spawn' is default if unused; also use 'forkserver' or 'fork' on Unix only

        # custom part, write your own Attributes
        self.range_num = 10 ** 4  # got a target/max value and NUM_ROWS for each proc, can calc ETA est. time arrival
        self.INFO_THREAD_MAX = self.range_num  # target value for info thread to calculate % and ETA
        # self.INFO_ENABLE = True  # [baustelle]
        self.n = 10 ** 12  # ten with zero count, 10_000_000_000_000_000
        self.say_hello = 'Hello'  # just to show that worker can [read] all attributes of instance, in 'worker_prime()'


modConf = ModuleConfiguration()  # Accessible in the module.


def generator_prime():
    """Manager - One time execution.
    Exits if generator is empty.
    """
    # auto
    mP = eisenmp.Mp()

    mP.start(**modConf.__dict__)  # instance attributes available for worker and feeder loop
    generator = number_generator()
    mP.run_q_feeder(generator=generator, input_q=mP.mp_input_q)  # here default mp_input_q, use if only one q needed

    return mP


def number_generator():
    """Generates numbers from start count.
    Has an end value, range.
    """
    range_num = modConf.range_num
    n = modConf.n  # start with a large number to see some work in progress
    for _ in range(range_num):
        yield n
        n += 1


def worker_prime(toolbox):
    """
    - WORKER - Called with a single arg (name it xyz) in a loop until returns False.

    Start, Entry, Exit of this 'single' process worker.
    We return True to get next list chunk, whatever object is in the rows.
    Fed from mp_input_q to our toolbox. toolbox is your work instance with queues,
    messages, list chunk, and work tools like language dictionary or hash list.

    toolbox.foo, gives also access to all attributes and values
    of the 'modConf.foo' instance, you have created
    """
    # toolbox.mp_print_q.put(toolbox.say_hello)
    busy = workload_get(toolbox)
    calc_prime(toolbox)  # start worker function, exit afterwards if stop received
    if not busy:
        return False
    send_eta_data(toolbox)  # send data list, first row is header, info thread can find it in eisenmp.output_q_box
    return True


def workload_get(toolbox):
    """"""
    while 1:
        if not toolbox.mp_input_q.empty():
            toolbox.NEXT_LIST = toolbox.mp_input_q.get()  # NEXT_LIST is pre-defined, you can declare your own var
            break
    if toolbox.STOP_MSG in toolbox.NEXT_LIST:  # eisenmp.iterator_loop() informs stop, no more lists
        return False  # loader sends shutdown msg to next worker - generator is empty
    return True


def calc_prime(toolbox):
    """Calc prime num.

    The stop message is detected in 'workload_get', in this example.
    """
    remove_header(toolbox)

    # calc
    lst = toolbox.NEXT_LIST  # NEXT_LIST is default var for the new list in mp_input_q
    stop_msg = toolbox.STOP_MSG  # string 'STOP'
    prime_lst = [str(num) for num in lst if num is not stop_msg and type(num) is int and is_prime(num)]

    send_output(toolbox, prime_lst)

    # print message
    primes = ''.join(str(prime_lst)) if len(prime_lst) else ''
    output_msg = f' ... Result {toolbox.WORKER_NAME} ... Prime {primes}'
    toolbox.mp_print_q.put(output_msg) if len(prime_lst) else None  # blocks the whole mp


def is_prime(n: int) -> bool:
    """https://en.wikipedia.org/wiki/Primality_test

    mod by 44xtc44, limit = int(math.sqrt(n)) , isqrt(n) is Python 3.8, package shall run 3.7
    """
    if n <= 3:
        return n > 1
    if n % 2 == 0 or n % 3 == 0:
        return False
    limit = int(math.sqrt(n))  # python 3.8 isqrt(n) humbug
    for i in range(5, limit + 1, 6):
        if n % i == 0 or n % (i + 2) == 0:
            return False
    return True


def send_output(toolbox, prime_lst):
    """Put your findings in the output list.
    Find results in the 'eisenmp_utils.Result.result_dict'

    :params: toolbox: -
    :params: prime_lst: findings of prime num calc
    """
    # OUTPUT_HEADER - mandatory, INPUT_HEADER - q collector distinguish queues, write own strings if multiple queues
    header = toolbox.OUTPUT_HEADER + toolbox.INPUT_HEADER  # INPUT_HEADER pre-defined, or write your own to find result
    result_lst = [header,
                  prime_lst]  # your findings here
    toolbox.mp_output_q.put(result_lst)


def send_eta_data(toolbox):
    """list of [PERF_HEADER_ETA, PERF_CURRENT_ETA] to ProcInfo, to calc arrival time ETA
    """
    toolbox.PERF_CURRENT_ETA = len(toolbox.NEXT_LIST)
    perf_lst = [toolbox.PERF_HEADER_ETA + toolbox.WORKER_NAME,  # binary head
                toolbox.PERF_CURRENT_ETA]
    # disable info q collector will block all
    toolbox.mp_info_q.put(perf_lst)  # ProcInfo calc arrival time and % from info_q, of all proc lists


def remove_header(toolbox):
    """Transport ticket with consecutive number.
    Remove if no recreation of order is necessary.
    """
    # toolbox.mp_print_q.put(toolbox.NEXT_LIST[0])
    toolbox.INPUT_HEADER = toolbox.NEXT_LIST[0]  # header with name of input Queue and serial number, for output Queue
    del toolbox.NEXT_LIST[0]  # remove header str from list, which you get from input Q


def main():
    """
    """
    start = time.perf_counter()

    mP = generator_prime()
    while 1:
        # running threads, wait
        if mP.begin_proc_shutdown:
            break
        time.sleep(1)

    msg_time = 'Example Prime numbers, Time in sec: ', round((time.perf_counter() - start))
    print(msg_time)
    msg_result = e_utils.Result.result_dict

    return msg_result


if __name__ == '__main__':
    main()



"""Module runs aacrepair on commandline with menu options.

Please check ``error_dict`` message, if instance fails to repair.
"""
import eisenmp_examples


def menu_main():
    """Main menu to choose from."""

    exit_msg = '\n  Thank you for using eisenmp_examples.'
    option_msg = 'Invalid option. Please enter a number between 1 and 8.'

    print('\n\tMenu "Example"\n')
    menu_options_dict = {
        1: 'Three processes, multiple Flask server in each - share a DB',
        2: 'Each CPU, one Flask server on each - share a DB',
        3: 'Prime Number calculation',
        4: 'Web CSV large list calculation',
        5: 'Each CPU, one simple http server presents a radio',
        6: 'Two Queues fed at once',
        7: 'Brute force attack with dictionary and itertools generator',
        8: 'Exit',
    }

    while 1:
        for key in menu_options_dict.keys():
            print(key, '--', menu_options_dict[key])
        try:
            option = int(input('Enter your choice: '))
        except ValueError:
            print(option_msg)
            continue
        if option == 1:
            eisenmp_examples.eisenmp_exa_multi_srv_each_cpu.main()
            break
        elif option == 2:
            eisenmp_examples.eisenmp_exa_each_flask_orm_srv_one_cpu.main()
            break
        elif option == 3:
            eisenmp_examples.eisenmp_exa_prime.main()
            break
        elif option == 4:
            eisenmp_examples.eisenmp_exa_web_csv.main()
            break
        elif option == 5:
            eisenmp_examples.eisenmp_exa_http.main()
            break
        elif option == 6:
            eisenmp_examples.eisenmp_exa_double_q.main()
            break
        elif option == 7:
            eisenmp_examples.eisenmp_exa_bruteforce.main()
            break

        elif option == 8:
            print(exit_msg)
            exit()
        else:
            print(option_msg)
    return


def main():
    """Call menu_main to start the module from command line."""
    menu_main()


if __name__ == '__main__':
    main()



"""Economy Example CSV column calculation.

"""


def worker_entrance(toolbox):
    """
    - WORKER - Called in a loop.

    Start, Entry, Exit of this single process worker.
    We return True to get next list chunk, whatever object is in the rows.
    Fed from mp_input_q to our toolbox. toolbox is our work instance with queues,
    messages, list chunk, and work tools like language dictionary or hash list.

    toolbox.foo, gives also access to all attributes and values
    of the 'modConf.foo' instance, you have created
    """
    # toolbox.mp_print_q.put(toolbox.say_hello)
    busy = workload_get(toolbox)
    calc_average(toolbox)  # start worker function
    if not busy:
        return False
    send_eta_data(toolbox)  # send data list, first row is header, info thread can find it in eisenmp.output_q_box
    return True


def workload_get(toolbox):
    """"""
    while 1:
        if not toolbox.mp_input_q.empty():
            toolbox.NEXT_LIST = toolbox.mp_input_q.get()
            break
    if toolbox.STOP_MSG in toolbox.NEXT_LIST:  # eisenmp.iterator_loop() informs stop, no more lists
        return False  # loop worker sends shutdown msg to next worker - generator is empty
    return True


def remove_header(toolbox):
    """Transport ticket with consecutive number.
    Remove if no recreation of order is necessary.
    Can reuse list for result, if rebuild order.

    Use self.header_msg attribute to overwrite default header string
    """
    # toolbox.mp_print_q.put(toolbox.NEXT_LIST[0])
    toolbox.INPUT_HEADER = toolbox.NEXT_LIST[0]
    del toolbox.NEXT_LIST[0]  # remove header str


def calc_average(toolbox):
    """Calc average from strings.
    Pandas can make float, but we use raw Python csv import.
    Table column has 'nan' and empty cells we can not read.
    """
    busy = True
    if toolbox.STOP_MSG in toolbox.NEXT_LIST:  # inform we want exit
        busy = False
    remove_header(toolbox)

    lst = toolbox.NEXT_LIST
    STOP_MSG = toolbox.STOP_MSG
    # kick out 'nan' string and binary stop message from list, stop message is appended on GhettoBoss iterator loop end
    tbl_flt = [float(num) for num in lst if str(num) and 'nan' not in str(num) and num != STOP_MSG]

    average = 0
    if len(tbl_flt):  # calc with float type to get comma values
        average = sum([num for num in tbl_flt]) / len(tbl_flt)
    average = average if average else 0

    send_output(toolbox, average)

    output_msg = f' ... {toolbox.WORKER_NAME} ... [ average |{average}| ] of {len(toolbox.NEXT_LIST)} rows'
    toolbox.mp_print_q.put(output_msg)

    return busy


def send_output(toolbox, average):
    """Put your findings in the output list.
    Find results in the 'eisenmp_utils.Result.result_dict'

    :params: toolbox: -
    :params: average: average of the (chunk of) column
    """
    # header for output result list
    header = toolbox.OUTPUT_HEADER + toolbox.INPUT_HEADER  # q collector can distinguish queues and store result in dict
    result_lst = [header,
                  average]  # your findings here
    toolbox.mp_output_q.put(result_lst)


def send_eta_data(toolbox):
    """list of [PERF_HEADER_ETA, PERF_CURRENT_ETA] to ProcInfo, to calc arrival time ETA
    """
    toolbox.PERF_CURRENT_ETA = len(toolbox.NEXT_LIST)
    perf_lst = [toolbox.PERF_HEADER_ETA + toolbox.WORKER_NAME,  # binary head
                toolbox.PERF_CURRENT_ETA]
    toolbox.mp_info_q.put(perf_lst)  # ProcInfo calc arrival time and % from mp_info_q, of all proc lists



import eisenmp_examples.eisenmp_exa_http
import eisenmp_examples.eisenmp_exa_prime
import eisenmp_examples.eisenmp_exa_web_csv
import eisenmp_examples.eisenmp_exa_double_q
import eisenmp_examples.eisenmp_exa_bruteforce
import eisenmp_examples.eisenmp_exa_multi_srv_each_cpu
import eisenmp_examples.eisenmp_exa_each_flask_orm_srv_one_cpu



# Python 3 server example
# https://pythonbasics.org/webserver/
# mod by 44xtc44

import os
import time
import platform
from http.server import BaseHTTPRequestHandler, HTTPServer

import eisenmp

hostName = "localhost"
serverPort = 12_000

radio_url_lst = [
    # just to show server is up, urls for entertainment, no practical use, see template for sqlAlchemy
    ('aac_chill         ', 'http://radio4.vip-radios.fm:8020/stream128k-AAC-Chill_autodj'),
    ('80ies_nl          ', 'http://94.228.133.5:80'),
    ('anime_jp          ', 'http://streamingv2.shoutcast.com/japanimradio-tokyo'),
    ('blues_uk          ', 'http://149.255.59.3:8232/stream'),
    ('classic_ro        ', 'http://37.251.146.169:8000/streamHD'),
    ('goa_psy           ', 'https://amoris.sknt.ru/goa.mp3'),
    ('hirsch_milch      ', 'https://hirschmilch.de:7001/prog-house.mp3'),
    ('nachtflug         ', 'http://85.195.88.149:11810'),
    ('paloma            ', 'https://pool.radiopaloma.de/RADIOPALOMA.mp3'),
    ('reggae            ', 'http://hd.lagrosseradio.info:8000/lagrosseradio-reggae-192.mp3'),
    ('Radio Time Machine', 'http://98.211.68.9:8765/;?icy=https'),
    ('yeah_mon          ', 'http://c3.radioboss.fm:8095/autodj'),
    ('zen_style         ', 'https://radio4.cdm-radio.com:18004/stream-mp3-Zen')
]


class ModuleConfiguration:
    """
    You can use the class to have your variables available in the module.

    'worker_http()' is executed by all processes on a CPU somewhere.
    Each 'worker_http()' gets (one by one) a list chunk from eisenmp.mp_input_q.

    """
    dir_name = os.path.dirname(__file__)  # our module path without file name
    # path to worker module and entry function reference, worker module import in [isolated] process environment
    # -------------------- MANDATORY WORKER STRINGS --------------------
    first_module = {
        'WORKER_PATH': os.path.join(dir_name, 'eisenmp_exa_http.py'),
        'WORKER_REF': 'worker_http',  # Warning: loader runs all f() with a single argument 'toolbox'; pull args from it
    }
    watchdog_module = {
        'WORKER_PATH': os.path.join(dir_name, 'worker', 'eisenmp_watchdog.py'),
        'WORKER_REF': 'mp_start_show_threads',
    }
    foo = {'WORKER_PATH': 'bar', 'WORKER_REF': 'baz'}

    def __init__(self):
        # load order list, first module is called in an endless loop, you can append your own loop inside the worker
        self.worker_modules = [
            self.first_module,  # second module must be threaded, else we hang
            # self.watchdog_module,  # enable for threaded module start live example
        ]

        # Multiprocess vars - override default
        self.NUM_PROCS = 16  # your process count, default is None: one proc/CPU core
        self.NUM_ROWS = 1  # workload spread, list (generator items) to calc in one loop, default is None: 1_000
        self.RESULTS_STORE = True  # keep in dictionary, will crash the system if store GB network chunks in mem
        self.RESULTS_PRINT = True  # result rows of output are collected in a list, display if processes are stopped
        self.RESULTS_DICT_PRINT = False  # shows content of results dict with ticket numbers, check tickets
        # max generator / NUM_ROWS = number of tickets, 10_000 / 42 = 238.095 -> 238 lists with ticket numbers
        self.RESULT_LABEL = 'revised.csv, Average calculation'  # pretty print as result header for RESULTS_PRINT
        if platform.system() == 'Linux':
            self.START_METHOD = 'fork'  # 'spawn' is default if unused; also use 'forkserver' or 'fork' on Unix only

        # 'not enough work example', useless worker auto shutdown, the modules return False
        self.radio_name = None  # define to get it as key in 'modConf' dictionary and worker use 'toolbox.radio_name'
        self.radio_url = None
        self.sleep_time = 60  # if watchdog enabled, toolbox.sleep_time = 60


modConf = ModuleConfiguration()  # Accessible in the module.


def manager_http_srv():
    """
    - Manager -

    """
    # default call feeds all variables of 'ModuleConfiguration' in kwargs -> toolbox of worker
    mP = eisenmp.Mp()
    mP.start(**modConf.__dict__)  # instance attributes available for worker and feeder loop
    # custom, don't need a generator and q feeder to run a server
    generator = (radio_url for radio_url in radio_url_lst)
    mP.run_q_feeder(generator=generator)


def worker_http(toolbox):  # arg for loader, all ModuleConfiguration instance vars can be pulled from toolbox
    """
    - Worker -

    Blocked, no loop here
    toolbox is the all-in-one box for vars and queues. incl. ModuleConfiguration
    We have num_procs, we have ports to add, we need worker in groups to serve ports
    """
    global serverPort

    WORKER_ID = toolbox.WORKER_ID
    serverPort = serverPort + WORKER_ID
    while 1:
        toolbox.next_lst = toolbox.mp_input_q.get()
        break
    if toolbox.STOP_MSG in toolbox.next_lst[1]:  # eisenmp.iterator_loop() informs (single list) stop, no more lists
        return False  # eisenmp_worker_loader will re-send the shutdown msg to the next worker - generator is empty

    toolbox.radio_name, toolbox.radio_url = toolbox.next_lst[1]  # list of tuples, [0] iterator list header
    MyServer.toolbox = toolbox
    webServer = HTTPServer((hostName, serverPort), MyServer)

    toolbox.mp_print_q.put(f"Server {WORKER_ID} started http://%s:%s" % (hostName, serverPort))
    try:
        webServer.serve_forever()
    except KeyboardInterrupt:
        pass
    webServer.server_close()
    print("Server stopped.")


class MyServer(BaseHTTPRequestHandler):
    toolbox = None

    def do_GET(self):  # no exception handling, else line is over 120, big enough for a basic, OMG an exception! Yep!
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()

        self.wfile.write(bytes("<!doctype html>", "utf-8"))
        self.wfile.write(bytes("<html><head>", "utf-8"))
        self.wfile.write(bytes("", "utf-8"))
        self.wfile.write(bytes("<meta charset='utf-8'>", "utf-8"))
        self.wfile.write(bytes("<meta name='viewport' content='width=device-width, ", "utf-8"))
        self.wfile.write(bytes("initial-scale=1, shrink-to-fit=no'/>", "utf-8"))
        self.wfile.write(bytes("<link rel=stylesheet href=https://fonts.googleapis.com/css?family=Roboto/>", "utf-8"))
        self.wfile.write(bytes("<title>https://pythonbasics.org</title>", "utf-8"))
        self.wfile.write(bytes("</head>", "utf-8"))
        self.wfile.write(bytes("<p>Request: %s</p>" % self.path, "utf-8"))
        self.wfile.write(bytes("<body>", "utf-8"))
        self.wfile.write(bytes("<h1 style='text-align:center;color:brown;font-family:Roboto;'> Server </h1> ", "utf-8"))
        self.wfile.write(bytes(f"<p style='text-align:center;color:brown;font-family:Roboto;'>", "utf-8"))
        self.wfile.write(bytes("Visit 44xtc44 on PyPi or SnapCraft.io, EisenRadio ,", "utf-8"))
        self.wfile.write(bytes("<a href=https://pypi.org/project/eisenradio/>link</a>", "utf-8"))
        self.wfile.write(bytes(" for a better Show, ", "utf-8"))
        self.wfile.write(bytes("Blacklist, recording and a 'Gain' slider\t", "utf-8"))
        self.wfile.write(bytes(f"\t{self.toolbox.radio_name}\n", "utf-8"))
        self.wfile.write(bytes(f"\t{self.toolbox.radio_url} <a href={self.toolbox.radio_url}>link</a>\n", "utf-8"))
        self.wfile.write(bytes("<hr>", "utf-8"))
        self.wfile.write(bytes("<div class='radio' style='text-align:center;font-family:Roboto;' >'", "utf-8"))
        self.wfile.write(bytes("<audio controls autoplay preload=metadata id='audioR' ", "utf-8"))
        self.wfile.write(bytes(f"src={self.toolbox.radio_url} ></audio> </div>", "utf-8"))
        self.wfile.write(bytes("<script>", "utf-8"))
        self.wfile.write(bytes("(function() {", "utf-8"))
        self.wfile.write(bytes("const audio = document.getElementById('audioR');audio.volume=0.5;", "utf-8"))
        self.wfile.write(bytes("})();", "utf-8"))
        self.wfile.write(bytes("</script>", "utf-8"))
        self.wfile.write(bytes("</body></html>", "utf-8"))
        self.wfile.write(bytes("", "utf-8"))

        self.toolbox.mp_print_q.put(f'proc name: {self.toolbox.WORKER_NAME} pid: {self.toolbox.WORKER_PID}')


def main():
    """
    """
    start = time.perf_counter()

    manager_http_srv()

    print('Time in sec: ', round((time.perf_counter() - start)))
    msg_time = 'HTTP Server, Time in sec: ', round((time.perf_counter() - start))
    print(msg_time)
    return msg_time


if __name__ == '__main__':
    main()



import os
import threading
from Flask_SQLAlchemy_Project_Template import create_app, setup_database, db_path

color_dict = {
    'PURPLE': '\033[1;35;48m',
    'CYAN': '\033[1;36;48m',
    'BOLD': '\033[1;37;48m',
    'BLUE': '\033[1;34;48m',
    'GREEN': '\033[1;32;48m',
    'YELLOW': '\033[1;33;48m',
    'yellow': '\033[1;31;48m',
    'BLACK': '\033[1;30;48m',
    'UNDERLINE': '\033[4;37;48m',
    'END': '\033[1;37;0m',
}

network = "localhost"


def worker(toolbox):
    """
    - Worker -

    """
    # port group
    proc_start_num = toolbox.kwargs['START_SEQUENCE_NUM']

    port, col = 0, None
    if proc_start_num in toolbox.blue_lst:
        col = color_dict['BLUE']
        port = blue_q_get(toolbox)[1]  # [0] is header row
    if proc_start_num in toolbox.yellow_lst:
        col = color_dict['YELLOW']
        port = yellow_q_get(toolbox)[1]
    if proc_start_num in toolbox.green_lst:
        col = color_dict['GREEN']
        port = green_q_get(toolbox)[1]

    col_end = color_dict['END']

    # Flask
    app_factory = create_app(port)  # flask, we feed port number to update the route -> Html page with our address
    if not os.path.isfile(db_path):  # do not kill db, if exists; MUST exist if many srv, else create by many srv, crash
        setup_database(app_factory)
    # app_factory.run(host="localhost", port=port)
    threading.Thread(
        target=lambda: app_factory.run(host="localhost", port=port)).start()

    msg = col + f'\nWORKER_MSG worker: {toolbox.WORKER_ID} pid: {toolbox.WORKER_PID} server port: {port}\n' \
                f'SERVER: http://{network}:{port}' + col_end
    toolbox.mp_print_q.put(msg)

    # end, return None (Nothing is None), loader leaves worker loop and waits for stop msg in mp_process_q


def blue_q_get(toolbox):
    """"""
    while 1:
        if not toolbox.mp_blue_q.empty():
            port_lst = toolbox.mp_blue_q.get()  # has header with serial number
            return port_lst


def yellow_q_get(toolbox):
    """"""
    while 1:
        if not toolbox.mp_yellow_q.empty():
            port_lst = toolbox.mp_yellow_q.get()
            return port_lst


def green_q_get(toolbox):
    """"""
    while 1:
        if not toolbox.mp_green_q.empty():
            port_lst = toolbox.mp_green_q.get()
            return port_lst



"""Compare string characteristics with 'opponent' str.
A timer shows periodically the amount of rows left.

"""

import time
from collections import defaultdict  # this time factory int to sum

try:
    import eisenmp.utils_exa.eisenmp_utils as e_utils
except ImportError:
    import eisenmp_examples.utils_exa.eisenmp_utils as e_utils


def worker_entrance(toolbox):
    """
    - WORKER - Called in a loop.

    Start, Entry, Exit of this single process worker.
    We return True to get next list chunk, whatever object is in the rows.
    Fed from mp_input_q to our toolbox. toolbox is our work instance with queues,
    messages, list chunk, and work tools like language dictionary or hash list.

    toolbox.foo, gives also access to all attributes and values
    of the 'modConf.foo' instance, you have created
    """
    # don't need a dictionary in mp_tools_q, we get generator lists to reduce, find a valid word
    busy = workload_get(toolbox)
    list_reduce(toolbox)  # start worker function
    if not busy:
        return False
    # send_eta_data(toolbox)  # no data, pb list shrink, worker has timer
    return True


def workload_get(toolbox):
    """"""
    while 1:
        if not toolbox.mp_input_q.empty():
            toolbox.NEXT_LIST = toolbox.mp_input_q.get()
            toolbox.num_lists += 1
            break
    if toolbox.STOP_MSG in toolbox.NEXT_LIST:  # eisenmp.iterator_loop() informs stop, no more lists
        return False  # loop worker sends shutdown msg to next worker - generator is empty
    return True


def remove_header(toolbox):
    """Transport ticket with consecutive number.
    Remove if no recreation of order is necessary.
    Can reuse list for result, if rebuild order.

    Use self.header_msg attribute to overwrite default header string
    """
    # toolbox.mp_print_q.put(toolbox.NEXT_LIST[0])
    toolbox.INPUT_HEADER = toolbox.NEXT_LIST[0]
    del toolbox.NEXT_LIST[0]  # remove header str


def send_output(toolbox, word_list):
    """Put your findings in the output list.
    Find results in the 'eisenmp_utils.Result.result_dict'

    :params: toolbox: -
    :params: average: average of the (chunk of) column
    """
    # header for output result list
    header = toolbox.OUTPUT_HEADER + toolbox.INPUT_HEADER  # q collector can distinguish queues and store result in dict
    result_lst = [header,
                  word_list]  # your findings here
    toolbox.mp_output_q.put(result_lst)


def list_reduce(toolbox):
    """Each Process takes only one chunk of the dict/list.
    Store char type [r, o, y] and count of search string.
    Compare string characteristics with 'opponent' str.
    Remove a word from list if char not in word or char type count does not match.
    Keep only matching words in the list.

    :params: str_to_comp: given str to search in wordlists/dict
    """
    busy = True
    if toolbox.STOP_MSG in toolbox.NEXT_LIST:  # inform we want exit
        busy = False
    remove_header(toolbox)

    str_to_comp = toolbox.str_permutation
    shrink_list = condense_list_from_mem_replace_spec_char(str_to_comp, toolbox.NEXT_LIST)
    word_list = shrink_list.copy()  # iter one and del other
    comp_typ_cnt_dict = char_type_count(str_to_comp)  # str compare

    start_time = time.perf_counter()
    for idx, word in enumerate(shrink_list):

        word_typ_cnt_dict = char_type_count(word)  # word in list to compare
        start_time = timer_show_rows_left(idx, start_time, word_list, toolbox)

        for char, char_count in comp_typ_cnt_dict.items():
            if char not in word_typ_cnt_dict.keys():
                if word in word_list:
                    word_list.remove(word)
            else:
                if word in word_list:
                    word_list.remove(word) if char_count != word_typ_cnt_dict[char] else None  # short circuit None last

    send_output(toolbox, word_list)

    return busy


def char_type_count(a_str):
    """Char count to identify string 'opponents' with unequal char count, but same length.
    """
    c_dict = defaultdict(int)
    for char in a_str:
        c_dict[char] += 1  # {'r': 3, 'o': 1, 'c': 1}
    return c_dict


def condense_list_from_mem_replace_spec_char(str_to_comp, word_list):
    """A list already in mem.
    Reduce. Replace str chars only for shrink candidates.

    :params: str_to_comp: search str
    :params: word_list: list to shrink with strings only of size str_to_comp
    """
    ret_list = []
    for word in word_list:
        if len(word) == len(str_to_comp):
            replaced = e_utils.replace_special_char(word)
            ret_list.append(replaced)
    return ret_list


def timer_show_rows_left(idx, start_time, word_list, toolbox):
    """Better readable.
    Show rows left in list.
    """
    seconds = 5
    if not (idx % 1_000):  # reduce performance hit, if any
        end_time = int(round((time.perf_counter() - start_time)))
        if end_time >= seconds:
            toolbox.mp_print_q.put(f'{toolbox.WORKER_NAME}. ... {len(word_list):,} rows, list({toolbox.num_lists}) ')
            start_time = time.perf_counter()
    return start_time  # kept if



"""Template for Manager

- Watchdog Thread included, shows pid of its proc
- double feeder
- custom queues and queues in category
- fake production of audio and video coding, is loop through to result dict
- two processors work on different streams (src is same here, so what)

"""
import os
import threading
import time
import platform
import eisenmp
import eisenmp.utils.eisenmp_utils as e_utils

dir_name = os.path.dirname(__file__)


chunks_0 = [
    # some 'binaries'
    b'\x94\x80\x12\'',
    bytes("<html><head>", "utf-8"),
    bytes(5),
    b'\x8e|\x1e\x95\'',
    b'\xbe\xb0-\xeex\'',
    b'\xf5\x98\x16$\'',
    b'\x13\xb7\x12XB\'',
    b'\x16\xbb\xe5MX\'',
    b'\x94\x80\x12\'',
    bytes("<html><head>", "utf-8"),
    bytes(5),
    b'\x8e|\x1e\x95\'',
    b'\xbe\xb0-\xeex\'',
    b'\xf5\x98\x16$\'',
    b'\x13\xb7\x12XB\'',
    b'\x16\xbb\xe5MX\'',
    b'\x94\x80\x12\'',
]

chunks_1 = [
    # some 'binaries'
    b'\x94\x80\x12\'',
    bytes("<html><head>", "utf-8"),
    b'\x13\xb7\x12XB\'',
    b'\x16\xbb\xe5MX\'',
    b'\x94\x80\x12\'',
    bytes(5),
    b'\x8e|\x1e\x95\'',
    b'\xbe\xb0-\xeex\'',
    b'\xf5\x98\x16$\'',
    b'\x13\xb7\x12XB\'',
    b'\x16\xbb\xe5MX\'',
    b'\x94\x80\x12\'',
    bytes("<html><head>", "utf-8"),
    b'\x13\xb7\x12XB\'',
    b'\x16\xbb\xe5MX\'',
    b'\x94\x80\x12\'',
    bytes(5),
    b'\x8e|\x1e\x95\'',
    b'\xbe\xb0-\xeex\'',
    b'\xf5\x98\x16$\'',
]


class ModuleConfiguration:  # name your own class and feed eisenmp with the dict
    """More advanced template. Multiprocess 'spawn' in 'ProcEnv' to work with all OS.
    - toolbox.kwargs shows all avail. vars and dead references of dicts, lists, instances, read only

    """
    template_module = {
        'WORKER_PATH': os.path.join(dir_name, 'worker', 'eisenmp_exa_wrk_double.py'),
        'WORKER_REF': 'worker_entrance',
    }
    watchdog_module = {
        'WORKER_PATH': os.path.join(os.path.dirname(dir_name), 'worker', 'eisenmp_exa_wrk_watchdog.py'),
        'WORKER_REF': 'mp_start_show_threads',
    }

    def __init__(self):
        super().__init__()
        self.worker_modules = [  # in-bld-res
            self.template_module,  # other modules must start threaded, else we hang
            self.watchdog_module  # second; thread function call mandatory, last module loaded first
        ]

        # Multiprocess vars - override default
        self.NUM_PROCS = 2  # your process count, default is None: one proc/CPU core
        self.NUM_ROWS = 3  # arbitrary num here
        self.RESULTS_STORE = True  # keep in dictionary, will crash the system if store GB network chunks in mem
        self.RESULTS_PRINT = True  # result rows of output are collected in a list, display if processes are stopped
        self.RESULT_LABEL = 'fake production of audio and video coding for WHO studios'
        self.RESULTS_DICT_PRINT = True  # shows content of results dict with ticket numbers, check tickets
        if platform.system() == 'Linux':
            self.START_METHOD = 'fork'  # 'spawn' is default if unused; also use 'forkserver' or 'fork' on Unix only

        # work to do
        self.sleep_time = 20  # watchdog
        self.num_of_lists = 0  # worker lists done counter

        # vars lived before in worker module classes
        self.header_aud = None  # pre-defined, use in whole worker mod, save input list header to mark output header
        self.header_vid = None  # output from multiple queues must be divided, default is only one output queue for all
        self.BLUE = '\033[1;34;48m'
        self.RED = '\033[1;31;48m'
        self.END = '\033[1;37;0m'


modConf = ModuleConfiguration()  # Accessible in the module.


def manager_entry():
    """
    - Generator - One time execution.

    Divide workload between processes / CPU
    -
    """
    q_cat_name_maxsize = [
        # q_category, q_name, q_maxsize; find your 100 Queues in the debugger, toolbox
        ('batch_1', 'audio_lg', 5),  # queues for batch_1
        ('batch_1', 'video_in', 1),  # dict avail. in worker module: toolbox.batch_1['video_in'].get()
        ('batch_7', 'audio_lg', 3),  # queues for batch_7, created but not used so far, can play with
        ('batch_7', 'video_in', 1),
    ]
    mP = eisenmp.Mp()

    # create custom queues with category and name
    mP.queue_cust_dict_category_create(*q_cat_name_maxsize)  # create queues, store in {custom} {category} dict

    audio_q_b1 = mP.queue_cust_dict_cat['batch_1']['audio_lg']  # USE Queue:
    video_q_b1 = mP.queue_cust_dict_cat['batch_1']['video_in']  # toolbox.batch_1['video_in'].get()
    audio_q_b7 = mP.queue_cust_dict_cat['batch_7']['audio_lg']
    video_q_b7 = mP.queue_cust_dict_cat['batch_7']['video_in']  # toolbox.batch_7['video_in'].get()

    mP.start(**modConf.__dict__)  # create processes, load worker mods, start threads (output_p coll, info)
    mP.run_q_feeder(generator=audio_generator_batch_1(), input_q=audio_q_b1)
    mP.run_q_feeder(generator=video_generator_batch_1(), input_q=video_q_b1)
    mP.run_q_feeder(generator=audio_generator_batch_7(), input_q=audio_q_b7)
    mP.run_q_feeder(generator=video_generator_batch_7(), input_q=video_q_b7)

    return mP


def audio_generator_batch_1():
    """"""
    for _ in chunks_0:
        yield _


def video_generator_batch_1():
    """"""
    for _ in chunks_1:
        yield _


def audio_generator_batch_7():
    """"""
    for _ in chunks_0:
        yield _


def video_generator_batch_7():
    """"""
    for _ in chunks_1:
        yield _


def main():
    """
    """
    start = time.perf_counter()

    mP = manager_entry()
    while 1:
        # running threads, wait
        if mP.begin_proc_shutdown:
            break
        time.sleep(1)

    msg_time = 'Fake production, Time in sec: ', round((time.perf_counter() - start))
    print(msg_time)
    msg_result = e_utils.Result.result_dict

    names_list = [thread.name for thread in threading.enumerate()]
    print(names_list)
    return msg_result


if __name__ == '__main__':
    main()



"""Generator module, Worker in /worker folder

Mandatory:
- import eisenmp
- worker must have the 'toolbox' as arg in entry function (arg count 1)
- path to worker module and the entry function reference, all are str

"""
import os
import io
import time
import platform
from zipfile import ZipFile

import eisenmp
import eisenmp.utils.eisenmp_utils as e_utils

try:
    from eisenmp.utils_exa.eisenmp_search import SearchStr
    from eisenmp.utils_exa.eisenmp_download import DownLoad
except ImportError:
    import eisenmp_examples
    from eisenmp_examples.utils_exa.eisenmp_search import SearchStr
    from eisenmp_examples.utils_exa.eisenmp_download import DownLoad

dir_name = os.path.dirname(__file__)  # absolute dir path


class ModuleConfiguration:
    """
    - Loading the worker module has nothing to do with name resolution in 'this' module.
    We can load from network share.

    We have full access to all queues and methods. eisenmp = eisenmp.Mp()

    """
    # path to worker module and entry function reference, worker module import in [isolated] process environment
    # -------------------- MANDATORY WORKER STRINGS (large worker live in /worker) --------------------
    bruteforce_module = {
        'worker_path': os.path.join(dir_name, 'worker', 'eisenmp_exa_wrk_bf_bruteforce.py'),
        'worker_ref': 'worker_entrance',
    }
    list_reducer_module = {
        'worker_path': os.path.join(dir_name, 'worker', 'eisenmp_exa_wrk_bf_reduce.py'),
        'worker_ref': 'worker_entrance',
    }
    lang_dicts_paths = [
        r'.\lang_dictionaries\ger\german.dic',
        r'.\lang_dictionaries\eng\words.txt'
    ]

    lang_dicts_url_zip_dct = {

        'https://github.com/44xtc44/eisenmp_examples/raw/dev/eisenmp_examples/lang_dictionaries/ger/german.zip':
            'german.dic',
        'https://github.com/44xtc44/eisenmp_examples/raw/dev/eisenmp_examples/lang_dictionaries/eng/SCOWL-wl.zip':
            'words.txt',
    }

    def __init__(self):
        # load order list, first module is called in an endless loop, you can append your own loop inside the worker
        self.worker_modules = []  # init for kwargs/toolbox, 'worker_module_set'

        # Multiprocess vars - override default
        # self.NUM_PROCS = 2  # your process count, default is None: one proc/CPU core
        # max generator / NUM_ROWS = number of tickets, 10_000 / 42 = 238.095 -> 238 lists with ticket numbers
        # self.NUM_ROWS = 2  # your workload spread, list (generator items) to calc in one loop, default is None: 1_000
        self.RESULTS_STORE = True  # keep in dictionary, will crash the system if store GB network chunks in mem
        self.RESULTS_PRINT = True  # result rows of output are collected in a list, display if processes are stopped
        self.RESULTS_DICT_PRINT = True  # shows content of results dict with ticket numbers, check tickets
        if platform.system() == 'Linux':
            self.START_METHOD = 'fork'  # 'spawn' is default if unused; also use 'forkserver' or 'fork' on Unix only

        # Worker part - toolbox vars survive multiple worker calls
        self.multi_tool_get = None  # custom var to allow only one download of MULTI_TOOL; init later in this example
        self.MULTI_TOOL = None  # pre-defined toolbox.MULTI_TOOL can host the dict from mp_tools_q
        self.INFO_THREAD_MAX = None  # target value for info thread to calculate % and ETA if 'enable_info' set

        # custom
        self.use_file_system = False  # False: download and unzip in mem        ------------ SWITCH --------------------
        self.result_lbl = None  # set by caller
        self.num_lists = 0  # add one for each list done, print out worker
        self.lowercase = True
        self.url = None  # 'use_file_system' False, URL of csv file
        self.zipped_filename = None  # None: we have to loop over two; name of the uncompressed file in zip archive
        self.str_permutation = None  # the search string we want to find in the dictionary/wordlist


modConf = ModuleConfiguration()  # accessible in module
searchStr = SearchStr()


def load_lang_word_dict():
    """Load dictionaries from disk or gitHub"""
    word_dicts_file_system_load() if modConf.use_file_system else word_dicts_git_hub_load()


def mp_start_raid():
    """
    - Manager -

    Decide raid tactics.
    Brute force on smaller strings. List reduce attempt for larger strings.

    Grab config values and push it to the two possible functions.

    - (A) len(String) <=  10, combined brute force dictionary attack.
    - (B) len(String) >=  11, Reduce a dictionary len(str) condensed list and count characters.
    """
    pass
    modConf.str_permutation = searchStr.search_string
    if modConf.lowercase:
        modConf.str_permutation = modConf.str_permutation.lower()
    searchStr.create_key_word_val_none_shrink(lowercase=modConf.lowercase)  # dict, remove words != len(search str)
    modConf.INFO_THREAD_MAX = len(searchStr.words_dict)  # info thread calc rows done and len

    # ---------- selection of generator function reference ----------
    brute_force = True if len(modConf.str_permutation) <= 10 else False
    function_ref = mp_brute_force if brute_force else mp_reduce  # decide which function to call
    worker_module_set(function_ref)

    msg_b, msg_r = f'\n\t[BRUTE_FORCE]\t{modConf.str_permutation}', f'\n\t[LIST_REDUCTION]\t{modConf.str_permutation}'
    msg = msg_b if function_ref == mp_brute_force else msg_r
    print(msg)

    generator = function_ref(searchStr)
    words_dict = searchStr.words_dict if function_ref == mp_brute_force else None

    mP = eisenmp.Mp()
    mP.start(**modConf.__dict__)
    if brute_force:
        for _ in mP.proc_list:
            mP.mp_tools_q.put(words_dict)
    mP.run_q_feeder(generator=generator, input_q=mP.mp_input_q)


def worker_module_set(function_ref):
    """which worker module and function to load"""
    modConf.worker_modules = []  # del existing

    if function_ref == mp_brute_force:
        modConf.multi_tool_get = True
        modConf.worker_modules.append(modConf.bruteforce_module)
    else:
        modConf.multi_tool_get = False
        modConf.worker_modules.append(modConf.list_reducer_module)


def mp_brute_force(search_instance):
    """Generator Part

    - Worker 1

    Produce str permutations for len(str), if len(str) = 3 we have 3! = 6 permutations
    """
    generator = search_instance.generator(lowercase=modConf.lowercase)  # itertools
    return generator


def mp_reduce(search_instance):
    """
    - Worker 2

    We took all language word list and put em in a dict.
    Deleted all words with more or less characters.

    Worker will now reduce the list generated from dict.
    Remove all words not matching characters and
    character count in search str from list.
    Worker puts result in output q.
    """
    generator = (word for word in search_instance.words_dict)  # generator expression; pre reduced in searchString;
    return generator


def word_dicts_file_system_load():
    """
    - Entry ZIP FS

    """
    extract_zip_wordlists()
    searchStr.create_key_word_val_none_dict(modConf.lang_dicts_paths, lowercase=modConf.lowercase)


def extract_zip_wordlists():
    """File system unzip.
    """
    dicts_ger_zip = os.path.join(dir_name, 'lang_dictionaries', 'ger', 'german.zip')
    dicts_eng_zip = os.path.join(dir_name, 'lang_dictionaries', 'eng', 'SCOWL-wl.zip')
    zip_lst = [dicts_ger_zip, dicts_eng_zip]
    for zip_file in zip_lst:
        with ZipFile(zip_file, 'r') as un_zip:
            un_zip.extractall(os.path.dirname(zip_file))


def word_dicts_git_hub_load():
    """
    - Entry ZIP load URL

    """
    downloader = DownLoad()
    mem_word_lst = []

    for url, in_zip in modConf.lang_dicts_url_zip_dct.items():
        modConf.url = url
        modConf.zipped_filename = in_zip
        wordlist_download(downloader)  # stored the response object
        mem_word_lst.extend(wordlists_in_memory(downloader))  # download the zip archive, open one file, return txt file

    if modConf.lowercase:  # store the word list in searchStr instance with lowercase if set {'ethic': None}
        # .rstrip() else single chars, later in shrink method of searchStr
        searchStr.loader_words_dict = {word.lower().rstrip(): None for word in mem_word_lst}
    else:
        searchStr.loader_words_dict = {word.rstrip(): None for word in mem_word_lst}
    pass


def wordlists_in_memory(downloader):
    """Open zip archive and load one file 'zipped_filename'
    Return as text.
    """
    archive = downloader.unzip_mem()
    # text files are all utf-8 encoded, py try open cp1252 on my windows
    with io.TextIOWrapper(archive.open(modConf.zipped_filename, 'r'), encoding="UTF-8") as file:
        txt_lst = file.readlines()
    return txt_lst


def wordlist_download(downloader):
    """Store response object in downloader instance.
    We could show dots during download, or open in binary, or csv, or ....
    """
    downloader.url = modConf.url
    downloader.zipped_filename = modConf.zipped_filename
    print(downloader.url)
    downloader.load_url()


def main():
    """
    """
    str_list_alphabet_salad = [
        'AMGNO',
        'AACEFHKLMSS',
        'DIKKLOOR', 'BEEINNRST',
        'CEEHNNRST', 'EEHINORST',
        'EEEFFIKORRS', 'CFHHILORRS'
    ]

    start = time.perf_counter()

    res_coll_dct = {}
    load_lang_word_dict()
    for idx, string in enumerate(str_list_alphabet_salad):
        searchStr.search_string = string
        modConf.result_lbl = string
        modConf.lowercase = True
        mp_start_raid()
        msg_result = e_utils.Result.result_dict
        res_coll_dct[idx] = msg_result
        print(msg_result)
        time.sleep(.5)

    print(f'BF Time in sec: {round((time.perf_counter() - start))}')
    return res_coll_dct


if __name__ == '__main__':
    main()
